// Copyright (C) 2024 Simon Quigley <tsimonq2@ubuntu.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "person.h"
#include "archive.h"
#include "distribution.h"
#include "launchpad.h"
#include <iostream>
#include <nlohmann/json.hpp>

person::person()
    : name(""),
      realname(""),
      web_link(""),
      email(""),
      description(""),
      visibility(""),
      lp(nullptr) {}

person::~person() {}

std::optional<person> person::parse(const std::string& json_data) {
    person p;
    p.parse_json(json_data);
    return p;
}

void person::parse_json(const std::string& json_data) {
    try {
        auto data = nlohmann::json::parse(json_data);

        // Map JSON keys to lambda functions
        std::map<std::string, std::function<void(const nlohmann::json&)>> json_map = {
            {"name", [this](const nlohmann::json& val) { name = val.get<std::string>(); }},
            {"realname", [this](const nlohmann::json& val) { realname = val.get<std::string>(); }},
            {"web_link", [this](const nlohmann::json& val) { web_link = val.get<std::string>(); }},
            {"email", [this](const nlohmann::json& val) { email = val.get<std::string>(); }},
            {"description", [this](const nlohmann::json& val) { description = val.get<std::string>(); }},
            {"visibility", [this](const nlohmann::json& val) { visibility = val.get<std::string>(); }},
            {"self_link", [this](const nlohmann::json& val) { self_link = val.get<std::string>(); }},
            {"ppas_collection_link", [this](const nlohmann::json& val) { ppas_collection_link = val.get<std::string>(); }},
            {"preferred_email_address_link", [this](const nlohmann::json& val) { preferred_email_address_link = val.get<std::string>(); }}
        };

        // Process JSON keys dynamically
        for (auto& [key, value] : data.items()) {
            if (json_map.find(key) != json_map.end()) {
                json_map[key](value);
                json_map.erase(key);  // Optional: remove processed key
            }
        }
    } catch (const nlohmann::json::parse_error& e) {
        std::cerr << "JSON parse error in person::parse_json: " << e.what() << std::endl;
        throw;
    }
}

std::optional<archive> person::getPPAByName(distribution& distro, const std::string& name_) {
    std::string owner = name;
    std::string endpoint = "~" + lp->url_encode(owner);
    std::map<std::string, std::string> params = {
        {"distribution", distro.self_link},
        {"name", name_},
        {"ws.op", "getPPAByName"}
    };
    auto response = lp->api_get(endpoint, params);
    if (!response.has_value()) {
        return std::nullopt;
    }

    std::optional<archive> ppa_opt = archive::parse(response.value());
    if (!ppa_opt.has_value()) return std::nullopt;

    // Set lp in the returned archive
    ppa_opt->set_lp(lp);
    return ppa_opt;
}

void person::set_lp(launchpad* lp_ptr) {
    lp = lp_ptr;
}
